
class Translate:
	def __init__(self):
		nucs = 'acgt'
		codons = [a+b+c for a in nucs for b in nucs for c in nucs]
		amino_acids = 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV#Y+YSSSS*CWCLFLF'
		self.translate = dict(zip(codons, amino_acids))
		ambi = """aarK aayN acrT acyT ackT acmT acbT acvT acdT achT agrR agyS atyI atmI athI 
		carQ cayH ccrP ccyP cckP ccmP ccbP ccvP ccdP cchP cgrR cgyR cgkR cgmR cgbR cgvR cgdR 
		cghR ctrL ctyL ctkL ctmL ctbL ctvL ctdL cthL garE gayD gcrA gcyA gckA gcmA gcbA gcvA 
		gcdA gchA ggrG ggyG ggkG ggmG ggbG ggvG ggdG gghG gtrV gtyV gtkV gtmV gtbV gtvV gtdV 
		gthV tar* tayY tcrS tcyS tckS tcmS tcbS tcvS tcdS tchS tgyC ttrL ttyF tra* ytaL ytgL 
		ytrL mgaR mggR mgrR"""
		for item in ambi.split():
			self.translate[item[0:3]] = item[-1]
		self.amino_acids = sorted(set(amino_acids))

		self.wobble = {'gcc':'cgg', 'gct':'cgg', 'gca':'cgt', 'gcg':'cgt',
					   'aga':'tct', 'agg':'tct', 'cga':'gct', 'cgg':'gct', 'cgt':'gcg', 'cgc':'gcg',
					   'gac':'ctg', 'gat':'ctg', 
					   'aac':'ttg', 'aat':'ttg',
					   'tgc':'acg', 'tgt':'acg',
					   'gaa':'ctt', 'gag':'ctt',
					   'caa':'gtt', 'cag':'gtt',
					   'gga':'cct', 'ggg':'cct', 'ggc':'ccg', 'ggt':'ccg',
					   'cac':'gtg', 'cat':'gtg',
					   'ata':'tat', 'atc':'tag', 'att': 'tag',
					   'tta':'aat', 'ttg':'aat', 'cta':'gta', 'ctg':'gta', 'ctt':'gtg', 'ctc':'gtg',
					   'aaa':'ttt', 'aag':'ttt',
					   'atg':'tac',
					   'ttc':'aag', 'ttt':'aag',
					   'cca':'ggt', 'ccg':'ggt', 'cct':'ggg', 'ccc':'ggg',
					   'agc':'tcg', 'agt':'tcg', 'tca':'atg', 'tcg':'atg', 'tcc':'agg', 'tct':'agg',
					   'aca':'tgt', 'acg':'tgt', 'acc':'tgg', 'act':'tgg',
					   'tgg':'acc',
					   'tac':'atg', 'tat':'atg',
					   'gta':'cat', 'gtg':'cat', 'gtc':'cag', 'gtt':'cag',
					   'tag':'atc', 'taa':'att', 'tga':'act'
					   }

	def rev_comp(self, seq):
		seq_dict = {'a':'t','t':'a','g':'c','c':'g',
					'n':'n',
					'r':'y','y':'r','s':'s','w':'w','k':'m','m':'k',
					'b':'v','v':'b','d':'h','h':'d'}
		return "".join([seq_dict[base] for base in reversed(seq)])

	def codon(self, codon):
		if len(codon) == 3:
			return self.translate.get(codon.lower(), 'X')
		else:
			return ''

	def sequence(self, seq, strand):
		aa = ''
		if strand > 0:
			for i in range(0, len(seq), 3):
				aa += self.codon(seq[i:i+3])
			return aa
		else:
			for i in range(0, len(seq), 3):
				aa += self.codon(self.rev_comp(seq[i:i+3]))
			return aa[::-1]
	
	def counts(self, seq, strand):
		aa = self.sequence(seq, strand)
		return Counter(aa)

	def frequencies(self, seq, strand):
		counts = self.counts(seq, strand)
		total = sum(counts.values())
		for aa in counts:
			counts[aa] = counts[aa] / total
		return counts



import re
import sys
import textwrap

from genbank.codons import Last
from genbank.codons import Next
from genbank.codons import Codons
from genbank.feature import Feature
from genbank.translate import Translate

def rev_comp(dna):
	a = 'acgtrykmbvdh'
	b = 'tgcayrmkvbhd'
	tab = str.maketrans(a,b)
	return dna.translate(tab)[::-1]


class Locus(dict):
	def __init__(self, name=None, dna=''):
		self.name = name
		self.dna = dna.lower()
		#self.codons = dict()
		self.translate = Translate()
	
	def construct_feature(self):
		'''this method allows for a Feature class to be modified through inheritance in other code '''
		return Feature

	def seq(self, left, right):
		return self.dna[left-1 : right]

	def length(self):
		return len(self.dna)

	def gc_content(self):
		if not hasattr(self, "gc"):
			a = self.dna.count('a')
			c = self.dna.count('c')
			g = self.dna.count('g')
			t = self.dna.count('t')
			self.gc = (c+g) / (a+c+g+t)
		return self.gc

	def pcodon(self, codon):
		codon = codon.lower()
		seq = self.dna + rev_comp(self.dna)
		p = dict()
		p['a'] = seq.count('a') / len(seq)
		p['c'] = seq.count('c') / len(seq)
		p['g'] = seq.count('g') / len(seq)
		p['t'] = seq.count('t') / len(seq)
		return p[codon[0]] * p[codon[1]] * p[codon[2]]

	def rbs(self):
		for feature in self:
			if feature.type == 'CDS':
				if feature.strand > 0:
					start = feature.left()+2
					feature.tags['rbs'] = self.seq(start-30,start)
				else:
					start = feature.right()
					feature.tags['rbs'] = rev_comp(self.seq(start,start+30))
	
	def features(self, include=None, exclude=None):
		for feature in self:
			if not include or feature.type in include:
				yield feature

	def add_feature(self, key, strand, pairs):
		"""Add a feature to the factory."""
		feature = self.construct_feature()
		feature = feature(key, strand, pairs, self)
		if feature not in self:
			self[feature] = len(self)
		return feature

	def read_feature(self, line):
		"""Add a feature to the factory."""
		key = line.split()[0]
		#partial  = 'left' if '<' in line else ('right' if '>' in line else False)
		strand = -1 if 'complement' in line else 1
		#pairs = [pair.split('..') for pair in re.findall(r"<*\d+\.\.>*\d+", line)]
		#pairs = [map(int, pair.split('..')) for pair in re.findall(r"<?\d+\.{0,2}>?\d+", line.replace('<','').replace('>','') )]
		pairs = [ pair.split('..') for pair in re.findall(r"<?\d+\.{0,2}>?\d*", line) ]
		# this is for weird malformed features
		#if ',1)' in line:
		#	pairs.append(['1','1'])
		# tuplize the pairs
		pairs = tuple([tuple(pair) for pair in pairs])
		feature = self.add_feature(key, strand, pairs)
		return feature

	def gene_coverage(self):
		''' This calculates the protein coding gene coverage, which should be around 1 '''
		cbases = tbases = 0	
		for locus in self.values():
			dna = [False] * len(locus.dna)
			seen = dict()
			for feature in locus.features(include=['CDS']):
				for i in feature.codon_locations():
					dna[i-1] = True
			cbases += sum(dna)
			tbases += len(dna)
		return 3 * cbases / tbases

	def write(self, outfile=sys.stdout):
		outfile.write('LOCUS       ')
		outfile.write(self.name)
		outfile.write(str(len(self.dna)).rjust(10))
		outfile.write(' bp    DNA             UNK')
		outfile.write('\n')
		outfile.write('DEFINITION  ' + self.name + '\n')
		outfile.write('FEATURES             Location/Qualifiers\n')
		#outfile.write('     source          1..')
		#outfile.write(str(len(self.dna)))
		#outfile.write('\n')
		for feature in self:
			feature.write(outfile)
		outfile.write('ORIGIN')
		i = 0
		dna = textwrap.wrap(self.dna, 10)
		for block in dna:
			if(i%60 == 0):
				outfile.write('\n')
				outfile.write(str(i+1).rjust(9))
				outfile.write(' ')
				outfile.write(block.lower())
			else:
				outfile.write(' ')
				outfile.write(block.lower())
			i += 10
		outfile.write('\n')
		outfile.write('//')
		outfile.write('\n')

	def last(self, n, strand, codons):
		if isinstance(codons, str):
			codons = [codons]
		if strand < +1:
			codons = list(map(rev_comp, codons))
		for i in range(n, 0, -3):
			if self.dna[i:i+3] in codons:
				return i
		return None

	def next(self, n, strand, codons):
		if isinstance(codons, str):
			codons = [codons]
		if strand < +1:
			codons = list(map(rev_comp, codons))
		for i in range(n, self.length(), 3):
			#print(list(codons), self.dna[i:i+3])
			if self.dna[i:i+3] in codons:
				return i
		return None

	def nearest(self, n, strand, codons):
		_last = self.last(n,strand,codons)
		_next = self.next(n,strand,codons)
		if n - _last < _next - n:
			return _last
		else:
			return _next

	def distance(self, n, strand, codons):
		nearest = self.nearest(n, strand, codons)
		return n - nearest if nearest < n else nearest - n





#!/usr/bin/env python3
# encoding: utf-8

import os
#from distutils.core import setup, Extension
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils import sysconfig


#os.environ["CC"] = "g++"

sources = ""

module = Extension('genbank',
			language='c',
			extra_compile_args=[''],
			extra_link_args=[''],
			include_dirs=[
						 '.',
						 '...',
						 os.path.join(os.getcwd(), 'include'),
			],
			sources = ["./src/"+item for item in sources.split()]
			)

def readme():
	with open("README.md", "r") as fh:
		long_desc = fh.read()
	return long_desc

def get_version():
	with open("VERSION", 'r') as f:
		v = f.readline().strip()
		return v

def main():
	setup (
		name = 'genbank',
		version = get_version(),
		author = "Katelyn McNair",
		author_email = "deprekate@gmail.com",
		description = 'Code to work with Genbank files',
		long_description = readme(),
		long_description_content_type="text/markdown",
		url =  "https://github.com/deprekate/genbank",
		scripts=['test.py'],
		classifiers=[
			"Programming Language :: Python :: 3",
			"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
			"Operating System :: OS Independent",
		],
		python_requires='>3.5.2',
		packages=find_packages(),
		#install_requires=[''],
		#ext_modules = [module],
	)


if __name__ == "__main__":
	main()



import sys

from genbank.file import File

f = File(sys.argv[1])
for name,locus in f.items():
	print(vars(locus))
	print("-----Locus:", name, "-----")
	for feature in locus:
		print(feature)
		'''
		if feature.type_is('CDS'):
			for codon in feature.codons():
				print(codon)
		'''
		#print(feature.locus.gc_content())
	stops = ['taa','tag','tga']
	print(locus.last(59, +1, stops))
	print(locus.last(59, +1, 'taa'))






class Next:
	def __init__(self, n=None):
		self.next = n
	def __repr__(self):
		return "%s(%s)" % (self.__class__.__name__, self.next)

class Last:
	def __init__(self, n=None):
		self.last = n
	def __repr__(self):
		return "%s(%s)" % (self.__class__.__name__, self.last)

class Codons(dict):
	def abs(self, n, item):
		distance = abs(item - n) if item else float("+Inf")
		return distance

	def nearest(self, n, lst):
		nearest = lst[0] if self.abs(n, lst[0]) < self.abs(n, lst[1]) else lst[1]
		return nearest

	def nearest_start(self, n, strand):
		n = n - 1
		lst = [ self.last_start[n].last , self.next_start[n].next ] if strand=='+' else [ self.last_start_[n].last , self.next_start_[n].next ]
		return self.nearest(n, lst) + 1

	def nearest_stop(self, n, strand):
		n = n - 1
		lst = [ self.last_stop[n].last , self.next_stop[n].next ] if strand=='+' else [ self.last_stop_[n].last , self.next_stop_[n].next ]
		return self.nearest(n, lst) + 1

	def last_start(self, n, strand):
		n = n - 1
		return self.last_start[n]

	def __init__(self, dna):
		pass


	def ininti(self, codons):
		self.last_start = [None] * len(dna)
		self.last_stop  = [None] * len(dna)
		self.next_start = [None] * len(dna)
		self.next_stop  = [None] * len(dna)
		self.last_start_ = [None] * len(dna)
		self.last_stop_  = [None] * len(dna)
		self.next_start_ = [None] * len(dna)
		self.next_stop_  = [None] * len(dna)
	
		lstarts = {0: Last(-1), 1: Last(-1), 2: Last(-1) }
		lstops  = {0: Last(-1), 1: Last(-1), 2: Last(-1) }
		nstarts = {0: Next(), 1: Next(), 2: Next() }
		nstops  = {0: Next(), 1: Next(), 2: Next() }
		lstarts_ = {0: Last(-1), 1: Last(-1), 2: Last(-1) }
		lstops_  = {0: Last(-1), 1: Last(-1), 2: Last(-1) }
		nstarts_ = {0: Next(), 1: Next(), 2: Next() }
		nstops_  = {0: Next(), 1: Next(), 2: Next() }

		i = 0
		for i,_ in enumerate(dna):
			codon = dna[i:i+3]
			frame = i % 3

			self.next_start[i] = nstarts[frame]
			# STARTS
			if codon in ['atg','gtg','ttg']:
				# last
				lstarts[frame] = Last(i)
				# next
				nstarts[frame].next = i
				nstarts[frame] = Next(i)
			self.last_start[i] = lstarts[frame]
			#---------------------------------
			self.next_start_[i] = nstarts_[frame]
			if codon in ['cat','cac','caa']:
				# last
				lstarts_[frame] = Last(i)
				# next
				nstarts_[frame].next = i
				nstarts_[frame] = Next(i)
			self.last_start_[i] = lstarts_[frame]

			# STOPS
			self.next_stop[i] = nstops[frame]
			if codon in ['tag','tga','taa']:
				# last
				lstarts[frame] = Last(i)
				lstops[frame] = Last(i)
				# next
				nstops[frame].next = i
				nstops[frame] = Next(i)
			self.last_stop[i] = lstops[frame]
			#---------------------------------
			self.next_stop_[i] = nstops_[frame]
			if codon in ['cta','tca','tta']:
				# last
				lstarts_[frame] = Last(i)
				lstops_[frame] = Last(i)
				# next
				nstops_[frame].next = i
				nstops_[frame] = Next(i)
			self.last_stop_[i] = lstops_[frame]
		# for orfs that run off the ends
		for frame in [0,1,2]:
			print(i-frame)
			self.next_stop[i-frame].next = i + 3
			self.next_start[i-frame].next = i + 3
			self.next_stop_[i-frame].next = i + 3
			self.next_start_[i-frame].next = i + 3
		#for i, obj in enumerate(self.next_stop):
		#	print(i, obj)



